@@ -20,6 +20,7 @@ __all__ = [
     'gid_of_uid',
     'uid_of_user',
     'updatefile',
+    'get_password',
 ]
 
 
@@ -30,10 +31,12 @@ import signal
 import stat
 import time
 import glob
-
+import re
 import fcntl
 import pwd
 import grp
+import getpass
+import commands
 
 from getmailcore.exceptions import *
 
@@ -48,6 +51,8 @@ _bool_values = {
     'off'   : False,
     '0'     : False
 }
+osx_keychain_binary = '/usr/bin/security'
+
 
 #######################################
 def lock_file(file, locktype):
@@ -93,13 +98,13 @@ class updatefile(object):
         self.closed = False
         self.filename = filename
         self.tmpname = filename + '.tmp.%d' % os.getpid()
-	# If the target is a symlink, the rename-on-close semantics of this
-	# class would break the symlink, replacing it with the new file.
-	# Instead, follow the symlink here, and replace the target file on
-	# close.
-	while os.path.islink(filename):
-	    filename = os.path.join(os.path.dirname(filename),
-	                            os.readlink(filename))
+        # If the target is a symlink, the rename-on-close semantics of this
+        # class would break the symlink, replacing it with the new file.
+        # Instead, follow the symlink here, and replace the target file on
+        # close.
+        while os.path.islink(filename):
+            filename = os.path.join(os.path.dirname(filename),
+                                    os.readlink(filename))
         try:
             f = safe_open(self.tmpname, 'wb')
         except IOError, (code, msg):
@@ -123,6 +128,7 @@ class updatefile(object):
         if self.closed or not hasattr(self, 'file'):
             return
         self.file.flush()
+        os.fsync(self.file.fileno())
         self.file.close()
         os.rename(self.tmpname, self.filename)
         self.closed = True
@@ -158,7 +164,7 @@ class logfile(object):
             # Seek to end
             self.file.seek(0, 2)
             self.file.write(time.strftime(logtimeformat, time.localtime())
-                + ' ' + s.rstrip() + os.linesep)
+                            + ' ' + s.rstrip() + os.linesep)
             self.file.flush()
         finally:
             unlock_file(self.file, 'flock')
@@ -249,8 +255,10 @@ def deliver_maildir(maildirpath, data, hostname, dcount=None, filemode=0600):
         if info['deliverycount'] is not None:
             info['unique'] += 'Q%(deliverycount)s' % info
         try:
-            info['unique'] += 'R%s' % ''.join(['%02x' % ord(char)
-                for char in open('/dev/urandom', 'rb').read(8)])
+            info['unique'] += 'R%s' % ''.join(
+                ['%02x' % ord(char)
+                 for char in open('/dev/urandom', 'rb').read(8)]
+            )
         except StandardError:
             pass
 
@@ -294,7 +302,7 @@ def deliver_maildir(maildirpath, data, hostname, dcount=None, filemode=0600):
     except IOError, o:
         signal.alarm(0)
         raise getmailDeliveryError('failure writing file %s (%s)'
-            % (fname_tmp, o))
+                                   % (fname_tmp, o))
 
     # Move message file from Maildir/tmp to Maildir/new
     try:
@@ -310,7 +318,7 @@ def deliver_maildir(maildirpath, data, hostname, dcount=None, filemode=0600):
         except StandardError:
             pass
         raise getmailDeliveryError('failure renaming "%s" to "%s"'
-            % (fname_tmp, fname_new))
+                                   % (fname_tmp, fname_new))
 
     # Delivery done
 
@@ -340,8 +348,10 @@ def eval_bool(s):
     try:
         return _bool_values[str(s).lower()]
     except KeyError:
-        raise getmailConfigurationError('boolean parameter requires value'
-            ' to be one of true or false, not "%s"' % s)
+        raise getmailConfigurationError(
+            'boolean parameter requires value to be one of true or false, '
+            'not "%s"' % s
+        )
 
 #######################################
 def gid_of_uid(uid):
@@ -398,7 +408,7 @@ def change_uidgid(logger=None, uid=None, gid=None):
                 os.setreuid(uid, uid)
     except OSError, o:
         raise getmailDeliveryError('change UID/GID to %s/%s failed (%s)'
-            % (uid, gid, o))
+                                   % (uid, gid, o))
 
 #######################################
 def format_header(name, line):
@@ -406,7 +416,7 @@ def format_header(name, line):
     '''
     header = ''
     line = (name.strip() + ': '
-        + ' '.join([part.strip() for part in line.splitlines()]))
+            + ' '.join([part.strip() for part in line.splitlines()]))
     # Split into lines of maximum 78 characters long plus newline, if
     # possible.  A long line may result if no space characters are present.
     while line and len(line) > 78:
@@ -452,12 +462,56 @@ def check_ssl_key_and_cert(conf):
     if certfile is not None:
         certfile = expand_user_vars(certfile)
     if keyfile and not os.path.isfile(keyfile):
-        raise getmailConfigurationError('optional keyfile must be'
-            ' path to a valid file')
+        raise getmailConfigurationError(
+            'optional keyfile must be path to a valid file'
+        )
     if certfile and not os.path.isfile(certfile):
-        raise getmailConfigurationError('optional certfile must be'
-            ' path to a valid file')
+        raise getmailConfigurationError(
+            'optional certfile must be path to a valid file'
+        )
     if (keyfile is None) ^ (certfile is None):
-        raise getmailConfigurationError('optional certfile and keyfile'
-            ' must be supplied together')
-    return keyfile, certfile
+        raise getmailConfigurationError(
+            'optional certfile and keyfile must be supplied together'
+        )
+    return (keyfile, certfile)
+
+
+#######################################
+if os.name == 'posix' and os.path.isfile(osx_keychain_binary):
+    def keychain_password(user, server, protocol, logger):
+        """Mac OSX: return a keychain password, if it exists.  Otherwise, return
+        None.
+        """
+        # wish we could pass along a comment to this thing for the user prompt
+        cmd = "%s find-internet-password -g -a '%s' -s '%s' -r '%s'" % (
+            osx_keychain_binary, user, server, protocol
+        )
+        (status, output) = commands.getstatusoutput(cmd)
+        if status != os.EX_OK or not output:
+            logger.error('keychain command %s failed: %s %s' 
+                         % (cmd, status, output))
+            return None
+        password = None
+        for line in output.split('\n'):
+            match = re.match(r'password: "([^"]+)"', line)
+            if match:
+                password = match.group(1)
+        if password is None:
+            logger.debug('No keychain password found for %s %s %s'
+                         % (user, server, protocol))
+        return password
+else:
+    def keychain_password(user, server, protocol, logger):
+        """Not Mac OSX: always return None.
+        """
+        return None
+
+
+#######################################
+def get_password(label, user, server, protocol, logger):
+    # try keychain first
+    password = keychain_password(user, server, protocol, logger)
+    # if no password found (or not on OSX), prompt in the usual way
+    if not password:
+        password = getpass.getpass('Enter password for %s:  ' % label)
+    return password
